/*
 * MiTCR <http://milaboratory.com>
 *
 * Copyright (c) 2010-2013:
 *     Bolotin Dmitriy     <bolotin.dmitriy@gmail.com>
 *     Chudakov Dmitriy    <chudakovdm@mail.ru>
 *
 * MiTCR is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation, either version 3 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>.
 */
package com.milaboratory.core.sequence;

import com.milaboratory.core.sequence.nucleotide.NucleotideSequence;
import com.milaboratory.core.sequence.nucleotide.NucleotideSequenceImpl;
import com.milaboratory.core.sequence.quality.SequenceQualityPhred;
import com.milaboratory.core.sequence.util.NucleotideRCSequence;
import com.milaboratory.core.sequence.util.NucleotideSubSequence;
import com.milaboratory.core.sequence.util.SequenceQualityUtils;
import com.milaboratory.core.sequence.util.SubSequenceQuality;

import java.io.Serializable;

/**
 * Main container of information about nucleotide sequence.
 *
 * <p>Contains sequence and quality values for all nucleotides in it.</p>
 *
 * @author Bolotin Dmitriy (bolotin.dmitriy@gmail.com)
 * @author Shugay Mikhail (mikhail.shugay@gmail.com)
 */
//TODO docs
public final class NucleotideSQPair implements Serializable {
    private static final long serialVersionUID = 1L;

    private NucleotideSequence sequence;
    private SequenceQuality quality;

    public NucleotideSQPair(String sequence) {
        this(NucleotideSequenceImpl.fromSequence(sequence));
    }

    public NucleotideSQPair(String sequence, String quality) {
        this(NucleotideSequenceImpl.fromSequence(sequence),
                new SequenceQualityPhred(quality));
    }

    public NucleotideSQPair(NucleotideSequence sequence, SequenceQuality quality) {
        if (sequence.size() != quality.size())
            throw new IllegalArgumentException();
        this.sequence = sequence;
        this.quality = quality;
    }

    public NucleotideSQPair(NucleotideSequence sequence) {
        this.sequence = sequence;
        this.quality = SequenceQualityUtils.getGoodQualityObject(sequence.size());
    }

    /**
     * Generates a new instance of NucleotideSQPair containing sub sequence. If to &lt; from then reverse complement
     * will be returned.
     *
     * @param from inclusive
     * @param to   exclusive
     */
    public NucleotideSQPair getSubPairCopy(int from, int to) {
        if (from >= size() || from < 0 || to > size() || to < 0)
            throw new IndexOutOfBoundsException("From or(and) to are out of range.");
        NucleotideSequence s;
        SequenceQuality q;
        if (to < from) {
            s = new NucleotideRCSequence(new NucleotideSubSequence(sequence, to + 1, from - to));
            q = new SubSequenceQuality(quality, to + 1, from - to).reverse();
        } else {
            s = new NucleotideSubSequence(sequence, from, to - from);
            q = new SubSequenceQuality(quality, from, to - from);
        }
        return new NucleotideSQPair(s, q);
    }

    public NucleotideSQPair getRC() {
        return new NucleotideSQPair(new NucleotideRCSequence(sequence), quality.reverse());
    }

    public NucleotideSQPair getDeepCopy() {
        return new NucleotideSQPair(NucleotideSequenceImpl.fromSequence(sequence), SequenceQualityUtils.copy(quality));
    }

    public int size() {
        return sequence.size();
    }

    public NucleotideSequence getSequence() {
        return sequence;
    }

    public SequenceQuality getQuality() {
        return quality;
    }

    public static String toPrettyString(NucleotideSQPair pair) {
        if (pair == null)
            return "null";
        String seq = pair.sequence.toString();
        String qual = pair.quality.toString();
        return seq + '\n' + qual;
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) return true;
        if (o == null || getClass() != o.getClass()) return false;

        NucleotideSQPair that = (NucleotideSQPair) o;

        if (!quality.equals(that.quality)) return false;
        if (!sequence.equals(that.sequence)) return false;

        return true;
    }

    @Override
    public int hashCode() {
        int result = sequence.hashCode();
        result = 31 * result + quality.hashCode();
        return result;
    }
}
