/*
 * MiTCR <http://milaboratory.com>
 *
 * Copyright (c) 2010-2013:
 *     Bolotin Dmitriy     <bolotin.dmitriy@gmail.com>
 *     Chudakov Dmitriy    <chudakovdm@mail.ru>
 *
 * MiTCR is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation, either version 3 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>.
 */
package com.milaboratory.core.sequence;


/**
 * Parent class for all sequence quality classes.
 *
 * @author Bolotin Dmitriy (bolotin.dmitriy@gmail.com)
 * @author Shugay Mikhail (mikhail.shugay@gmail.com)
 */
public interface SequenceQuality {
    Class<? extends SequenceQuality> getBaseClass();

    /**
     * Get the probability of error (e.g. nucleotide substitution) at given sequence point
     *
     * @param coord coordinate in sequence
     * @return probability of error
     */
    float probabilityOfErrorAt(int coord);

    /**
     * Get the log10 of probability of error (e.g. nucleotide substitution) at given sequence point
     *
     * @param coord coordinate in sequence
     * @return log10 of probability of error
     */
    float log10ProbabilityOfErrorAt(int coord);

    /**
     * Get the raw sequence quality value (in binary format) at given sequence point
     *
     * @param coord coordinate in sequence
     * @return raw sequence quality value
     */
    byte rawQualityValue(int coord);

    /**
     * Sets the raw sequence quality value (in binary format) at given sequence point
     *
     * @param coord coordinate in sequence
     */
    void setRawQualityValue(int coord, byte value);

    /**
     * Get the worst raw sequence quality value (in binary format) in all sequence
     *
     * @return worst raw sequence quality value
     */
    byte minRawQuality();

    /**
     * Merges two quality arrays by choosing the best quality at each position
     *
     * @param quality other quality array
     */
    void mergeWith(SequenceQuality quality);

    /**
     * Gets the qualities of characters in reverse order
     *
     * @return reverse of quality array
     */
    SequenceQuality reverse();

    /**
     * Gets the size of quality array
     *
     * @return size of quality array
     */
    int size();

    byte[] encode(byte offset);

    /**
     * Default quality thresholds
     */
    public static class Constants {
        public static final byte goodQuality = 34;
        public static final byte badQuality = 3;
    }
}
