/*
 * MiTCR <http://milaboratory.com>
 *
 * Copyright (c) 2010-2013:
 *     Bolotin Dmitriy     <bolotin.dmitriy@gmail.com>
 *     Chudakov Dmitriy    <chudakovdm@mail.ru>
 *
 * MiTCR is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation, either version 3 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>.
 */
package com.milaboratory.core.sequence.quality;

import com.milaboratory.core.sequence.SequenceQuality;

import java.io.Serializable;
import java.util.Arrays;

/**
 * Implementation sequence quality.
 *
 * <p>Phred sequence quality format used. It is common format for all new fastq formats (Sanger, Solexa v1.4+).</p>
 *
 * @author Bolotin Dmitriy (bolotin.dmitriy@gmail.com)
 * @author Shugay Mikhail (mikhail.shugay@gmail.com)
 */
public final class SequenceQualityPhred implements SequenceQuality, Serializable {
    private static final long serialVersionUID = 1L;

    private byte[] data;

    public SequenceQualityPhred(String string) {
        this(string, 33);
    }

    public SequenceQualityPhred(String string, int offset) {
        this.data = string.getBytes();
        for (int i = this.data.length - 1; i >= 0; --i)
            this.data[i] -= offset;
    }

    public SequenceQualityPhred(byte value, int length) {
        data = new byte[length];
        Arrays.fill(data, value);
    }

    public SequenceQualityPhred(byte[] data) {
        this.data = data;
    }

    @Override
    public Class<? extends SequenceQuality> getBaseClass() {
        return SequenceQualityPhred.class;
    }

    @Override
    public float log10ProbabilityOfErrorAt(int coord) {
        return -((float) data[coord]) / 10;
    }

    @Override
    public float probabilityOfErrorAt(int coord) {
        return (float) Math.pow(10.0, -(data[coord]) / 10);
    }

    @Override
    public byte rawQualityValue(int coord) {
        return data[coord];
    }

    @Override
    public void setRawQualityValue(int coord, byte value) {
        data[coord] = value;
    }

    @Override
    public byte minRawQuality() {
        byte min = Byte.MAX_VALUE;
        for (byte b : data)
            if (b < min)
                min = b;
        return min;
    }

    public byte[] getRawData() {
        return data;
    }

    public void set(int index, byte value) {
        data[index] = value;
    }

    @Override
    public SequenceQuality reverse() {
        return new SequenceQualityPhred(reverceTransformQualityStorage(data));
    }

    public static byte[] reverceTransformQualityStorage(byte[] quality) {
        byte[] newData = new byte[quality.length];
        int reverseCoord;
        for (int coord = 0; coord < quality.length; coord++) {
            reverseCoord = quality.length - 1 - coord;
            newData[coord] = quality[reverseCoord];
        }
        return newData;
    }

    @Override
    public void mergeWith(SequenceQuality quality) {
        if (quality.getBaseClass() != SequenceQualityPhred.class
                || quality.size() != this.data.length)
            return;
        byte q;
        for (int i = 0; i < data.length; ++i)
            if (data[i] < (q = quality.rawQualityValue(i)))
                data[i] = q;
    }

    @Override
    public int size() {
        return data.length;
    }

    @Override
    public String toString() {
        char[] chars = new char[data.length];
        for (int i = 0; i < data.length; ++i)
            chars[i] = (char) (33 + data[i]);
        return new String(chars);
    }

    public static byte[] getContent(SequenceQualityPhred qualityPhred) {
        return qualityPhred.data.clone();
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) return true;
        if (o == null || getClass() != o.getClass()) return false;

        SequenceQualityPhred that = (SequenceQualityPhred) o;

        if (!Arrays.equals(data, that.data)) return false;

        return true;
    }

    @Override
    public byte[] encode(byte offset) {
        byte[] copy = data.clone();
        for (int i = copy.length - 1; i >= 0; --i)
            copy[i] += offset;
        return copy;
    }

    @Override
    public int hashCode() {
        return Arrays.hashCode(data);
    }
}
