/*
 * MiTCR <http://milaboratory.com>
 *
 * Copyright (c) 2010-2013:
 *     Bolotin Dmitriy     <bolotin.dmitriy@gmail.com>
 *     Chudakov Dmitriy    <chudakovdm@mail.ru>
 *
 * MiTCR is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation, either version 3 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>.
 */
package com.milaboratory.mitcr.cli;

import com.milaboratory.core.clone.Clone;
import com.milaboratory.core.clone.CloneImpl;
import com.milaboratory.core.clone.CloneSet;
import com.milaboratory.core.clone.CloneSetImpl;
import com.milaboratory.core.segment.*;
import com.milaboratory.core.sequence.NucleotideSQPair;
import com.milaboratory.core.sequence.nucleotide.NucleotideSequenceImpl;
import com.milaboratory.core.sequencing.io.fastq.quality.QualityStringFormat;
import com.milaboratory.util.BitArray;

import java.io.*;
import java.util.ArrayList;
import java.util.List;

/**
 * 1/7/13 MS (mikhail.shugay@gmail.com)
 */
public final class TabDelimitedImporter {
    private TabDelimitedImporter() {
    }

    /**
     * Reads clone set formatted as tab-delimited table using {@link TabDelimitedExporter#exportCloneSet(String,
     * com.milaboratory.core.clone.CloneSet, ExportDetalizationLevel)} with detalization level equals to {@link
     * ExportDetalizationLevel#Full}.
     *
     * @param fileName file name
     * @return resulting clone set
     */
    public static CloneSet importCloneSet(String fileName) throws IOException {
        return importCloneSet(new FileInputStream(fileName));
    }

    public static CloneSet importCloneSet(InputStream stream) throws IOException {
        return importCloneSet(stream, DefaultSegmentLibrary.load());
    }

    public static CloneSet importCloneSet(InputStream stream, SegmentLibrary library) throws IOException {

        BufferedReader reader = new BufferedReader(new InputStreamReader(stream));

        String line = reader.readLine();

        //Checking first bytes
        if (!line.startsWith(TabDelimitedExporter.FULL_MAGIC_PREFIX))
            throw new IOException("Wrong file format.");

        String[] splitLine = line.split("\t");

        //Retrieving gene and species from header
        Species species;
        try {
            species = Species.valueOf(splitLine[1]);
        } catch (IllegalArgumentException iae) {
            throw new IOException("Wrong file format.");
        }
        Gene gene = Gene.fromXML(splitLine[2]);
        if (gene == null)
            throw new IOException("Wrong file format.");

        //Redundant check
        if ((splitLine = reader.readLine().split("\t")).length != 19)
            throw new IOException("Bad number of columns in header.");

        /*
            Format summary:

                0 clone.getCount(),
                1 clone.getPart(),
                2 clone.getCDR3().getSequence(),
                3 clone.getCDR3().getQuality(),
                4 clone.getCDR3().getQuality().minRawQuality(),
                5 clone.getCDR3AA(),
                6 clone.getVAlleles(),
                7 clone.getVSegments(),
                8 clone.getJAlleles(),
                9 clone.getJSegments(),
                10 clone.getDAlleles(),
                11 clone.getDSegments(),
                12 clone.getVEnd(),
                13 clone.getDStart(),
                14 clone.getDEnd(),
                15 clone.getJStart(),
                16 clone.insertionsVD(),
                17 clone.insertionsDJ(),
                18 clone.insertionsTotal()
        */


        //Std Sanger quality format with 33 offset
        final QualityStringFormat format = QualityStringFormat.Illumina18u;

        int id = 0;

        //SegmentLibrary library = DefaultSegmentLibrary.load();
        final SegmentGroupContainer vContainer = library.getGroup(species, gene, SegmentGroupType.Variable),
                jContainer = library.getGroup(species, gene, SegmentGroupType.Joining),
                dContainer = gene.hasDSegment() ? library.getGroup(species, gene, SegmentGroupType.Diversity) : null;

        List<Clone> clones = new ArrayList<>();
        CloneImpl clone;
        while ((line = reader.readLine()) != null) {
            splitLine = line.split("\t");

            //Retrieving reads count
            long count = Long.parseLong(splitLine[0]);

            //Retrieving segments for current clone
            BitArray vBarcode = new BitArray(vContainer.getAllelesCount());
            for (String vSegmentName : splitLine[6].split(","))
                vBarcode.set(vContainer.getAlleleByName(vSegmentName.trim()).getIndex());

            BitArray jBarcode = new BitArray(jContainer.getAllelesCount());
            for (String jSegmentName : splitLine[8].split(","))
                jBarcode.set(jContainer.getAlleleByName(jSegmentName.trim()).getIndex());

            BitArray dBarcode = null;
            if (dContainer != null) {
                dBarcode = new BitArray(dContainer.getAllelesCount());
                if (!splitLine[10].isEmpty())
                    for (String dSegment : splitLine[10].split(","))
                        dBarcode.set(dContainer.getAlleleByName(dSegment.trim()).getIndex());
            }
            BitArray[] segmentBarcodes = new BitArray[]{vBarcode, jBarcode, dBarcode};

            //Retrieving segments layout for current clone
            int[] segmentCoords;
            if (Integer.parseInt(splitLine[13]) > -1)
                segmentCoords = new int[]{Integer.parseInt(splitLine[12]), Integer.parseInt(splitLine[13]),
                        Integer.parseInt(splitLine[14]), Integer.parseInt(splitLine[15])};
            else
                segmentCoords = new int[]{Integer.parseInt(splitLine[12]), Integer.parseInt(splitLine[15])};

            clone = new CloneImpl(id++, new NucleotideSQPair(NucleotideSequenceImpl.fromSequence(splitLine[2]),
                    format.getQualityFactory().create(splitLine[3].getBytes())), count, segmentBarcodes, segmentCoords);
            clones.add(clone);
        }

        //Creating clone set
        return new CloneSetImpl<>(clones, gene, species, vContainer, jContainer, dContainer);
    }
}
