/*
 * MiTCR <http://milaboratory.com>
 *
 * Copyright (c) 2010-2013:
 *     Bolotin Dmitriy     <bolotin.dmitriy@gmail.com>
 *     Chudakov Dmitriy    <chudakovdm@mail.ru>
 *
 * MiTCR is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation, either version 3 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>.
 */
package com.milaboratory.mitcr.clonegenerator;

import com.milaboratory.core.clone.CloneImpl;
import com.milaboratory.core.sequence.NucleotideSQPair;
import com.milaboratory.core.sequencing.read.SequencingRead;
import com.milaboratory.mitcr.cdrextraction.CDR3ExtractionResult;
import com.milaboratory.mitcr.vdjmapping.SegmentMappingResult;
import com.milaboratory.util.BitArray;

import java.util.ArrayList;

/**
 * Object to collect information about {@link com.milaboratory.core.clone.Clone}. This class is thread-safe. Used
 * internally
 *
 * @author Bolotin Dmitriy (bolotin.dmitriy@gmail.com)
 * @author Shugay Mikhail (mikhail.shugay@gmail.com)
 */
final class AccumulatorClone extends CloneImpl {
    long additionalCount = 0;
    private BarcodeAggregator[] bcAgregators = new BarcodeAggregator[3];

    AccumulatorClone(int id, NucleotideSQPair cdr3, boolean saveLinks) {
        super(id, cdr3);
        if (saveLinks)
            links = new ArrayList<>();
        else
            links = null;
    }

    final void loadSegmentsCoordsFromMappingResult(CDR3ExtractionResult result) {
        int shift = result.getVMappingResult().getRefPoint() - 3;
        int vTo = result.getVMappingResult().getSegmentBorderTo() - shift;
        int jFrom = result.getJMappingResult().getSegmentBorderFrom() - shift;

        boolean tt = true;
        while (vTo >= jFrom)
            if (tt = !tt) //This is right :)) [Tick-tock pattern] 
                --vTo; //Tock 1, 3, 5, ...
            else
                ++jFrom; //Tick 0, 2, 4, ...

        final SegmentMappingResult dResult = result.getDMappingResult();
        //int dFrom = -1, dTo = -1;

        if (dResult != null) {
            segmentCoords = new int[4];
            segmentCoords[0] = vTo;
            segmentCoords[1] = dResult.getSegmentBorderFrom() - shift;
            segmentCoords[2] = dResult.getSegmentBorderTo() - shift;
            segmentCoords[3] = jFrom;
        } else {
            segmentCoords = new int[2];
            segmentCoords[0] = vTo;
            segmentCoords[1] = jFrom;
        }
    }

    private void addLink(CDR3ExtractionResult result, boolean additional) {
        //This operation is defined only for Long or SequencingRead sources
        //E.g. will do nothing for CDR3ExtractionResult created for NucleotideSQPair
        if (links != null &&
                (result.getSource() instanceof SequencingRead || result.getSource() instanceof Long)) {

            //Retrieving read id
            long id;
            if (result.getSource() instanceof SequencingRead)
                id = ((SequencingRead) result.getSource()).id();
            else
                id = (Long) result.getSource();

            //Adding link
            links.add(new SequencingReadLink(result.getVMappingResult().getRefPoint() - 3,
                    result.getJMappingResult().getRefPoint() + 3,
                    id, additional, result.isFoundInReverseComplement()));
        }
    }

    synchronized final void include(CDR3ExtractionResult result, boolean additional) {
        if (segmentCoords == null)
            loadSegmentsCoordsFromMappingResult(result);

        cdr3.getQuality().mergeWith(result.getCDR3().getQuality());
        BitArray barcode;
        for (int i = 0; i < 3; ++i) {
            barcode = result.getBarcode(i);
            if (barcode != null) {
                if (bcAgregators[i] == null)
                    bcAgregators[i] = new BarcodeAggregator(barcode.size());
                bcAgregators[i].addBarcode(barcode);
            }
        }
        addLink(result, additional);
        ++count;

        if (additional)
            ++additionalCount;
    }

    public int id() {
        return id;
    }

    void compile(float barcodeAggregationFactor) {
        for (int i = 0; i < 3; ++i)
            if (bcAgregators[i] == null)
                barcodes[i] = null;
            else
                barcodes[i] = bcAgregators[i].calculateBarcode(barcodeAggregationFactor);

        //Freeing memory
        this.bcAgregators = null;
    }
}
